from __future__ import annotations

from dataclasses import dataclass

INPUT_RANGE = (0.0, 1.0)
OUTPUT_RANGE = (0.0, 10.0)
ROUND_PRECISION = 1


@dataclass
class TriangularInterpolationOptions:
    """Options for triangular interpolation."""

    input_range: tuple[float, float] = INPUT_RANGE
    output_range: tuple[float, float] = OUTPUT_RANGE
    round_precision: int = ROUND_PRECISION


def triangular_interpolation(
    value: float, options: TriangularInterpolationOptions | None = None
) -> float:
    """
    Maps a value between input range to a triangular pattern between output range.
    Values between input_min-midpoint map from output_min to output_max linearly.
    Values between midpoint-input_max map from output_max to output_min linearly.
    The midpoint is automatically calculated as the middle of the input range.
    Result is always rounded to the specified precision.

    Args:
        value (float): Input value between input_min and input_max
        options (TriangularInterpolationOptions, optional): Configuration options for the interpolation.
            If not provided, default values will be used.

    Returns:
        float: Interpolated value between output_min and output_max, rounded to specified precision
    """
    if options is None:
        options = TriangularInterpolationOptions()

    # Extract values from options for readability
    input_min, input_max = options.input_range
    output_min, output_max = options.output_range
    round_precision = options.round_precision

    # Calculate midpoint
    midpoint = input_min + (input_max - input_min) / 2

    # Ensure input is between input_min and input_max
    clamped_value = max(input_min, min(input_max, value))

    if clamped_value <= midpoint:
        # First half: linear mapping from input_min→output_min to midpoint→output_max
        normalized_value = (clamped_value - input_min) / (midpoint - input_min)
        result = output_min + normalized_value * (output_max - output_min)
    else:
        # Second half: linear mapping from midpoint→output_max to input_max→output_min
        normalized_value = (clamped_value - midpoint) / (input_max - midpoint)
        result = output_max - normalized_value * (output_max - output_min)

    # Round to specified precision
    result = round(result, round_precision)

    return result


def param_safe_unique_id(length: int) -> str:
    """
    Generate a random string of alphabetic characters.

    Args:
        length (int): The length of the string to generate

    Returns:
        str: A random string of alphabetic characters
    """
    import random

    result = ""
    characters = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
    counter = 0
    while counter < length:
        result += characters[random.randint(0, len(characters) - 1)]  # noqa: S311 # doesn't need to be cryptographically secure
        counter += 1
    return result



from __future__ import annotations

import enum
from typing import NoReturn


def raise_with_error_code(e: Exception, error_code: ErrorCode) -> NoReturn:
    """Raise a GXCoreError with the given error_code and the message from the given exception.

    Meant to standardize the way we raise GXCoreErrors with error codes.
    """
    raise GXCoreError(message=str(e), error_code=error_code) from e


class GXCoreError(Exception):
    """To bridge between 0.18.x and 1.0 we create the GXCoreError base class in the agent code
    and add additional properties if relevant when subclassing. This should be moved to the
    GX Core repo when error codes are added to the core codebase in v1."""

    def __init__(self, message: str, error_code: str) -> None:
        super().__init__(message)
        self.error_code = error_code

    def get_error_params(self) -> dict[str, str]:
        """Return a dictionary of error parameters to be included in the error response,
        not including the error message or error code."""
        return {k: str(v) for k, v in vars(self).items() if k not in ["message", "error_code"]}


class ErrorCode(str, enum.Enum):
    """Error codes for GX Core errors."""

    GENERIC_UNHANDLED_ERROR = "generic-unhandled-error"
    WRONG_USERNAME_OR_PASSWORD = "wrong-username-or-password"  # noqa: S105  # Not a hardcoded password


class GXAgentError(Exception): ...


class GXAgentConfigError(GXAgentError): ...


class GXAgentUnrecoverableConnectionError(GXAgentError): ...



from __future__ import annotations

from urllib.parse import urljoin

from great_expectations.core.http import create_session
from great_expectations.datasource.fluent import SQLDatasource
from great_expectations.exceptions import GXCloudError
from typing_extensions import override

from great_expectations_cloud.agent.actions.agent_action import (
    ActionResult,
    AgentAction,
)
from great_expectations_cloud.agent.actions.utils import get_table_names
from great_expectations_cloud.agent.event_handler import register_event_action
from great_expectations_cloud.agent.models import (
    ListTableNamesEvent,
)


class ListTableNamesAction(AgentAction[ListTableNamesEvent]):
    # TODO: New actions need to be created that are compatible with GX v1 and registered for v1.
    #  This action is registered for v0, see register_event_action()

    @override
    def run(self, event: ListTableNamesEvent, id: str) -> ActionResult:
        datasource_name: str = event.datasource_name
        datasource = self._context.data_sources.get(name=datasource_name)
        if not isinstance(datasource, SQLDatasource):
            raise TypeError(  # noqa: TRY003 # one off error
                f"This operation requires a SQL Data Source but got {type(datasource).__name__}."
            )

        table_names = get_table_names(datasource)

        self._add_or_update_table_names_list(
            datasource_id=str(datasource.id), table_names=table_names
        )

        return ActionResult(
            id=id,
            type=event.type,
            created_resources=[],
        )

    def _add_or_update_table_names_list(self, datasource_id: str, table_names: list[str]) -> None:
        with create_session(access_token=self._auth_key) as session:
            url = urljoin(
                base=self._base_url,
                url=f"/api/v1/organizations/{self._organization_id}/table-names/{datasource_id}",
            )
            response = session.put(
                url=url,
                json={"data": {"table_names": table_names}},
            )
        if response.status_code != 200:  # noqa: PLR2004
            raise GXCloudError(
                message=f"ListTableNamesAction encountered an error while connecting to GX Cloud. "
                f"Unable to update "
                f"table_names for Data Source with ID"
                f"={datasource_id}.",
                response=response,
            )


register_event_action("1", ListTableNamesEvent, ListTableNamesAction)
